#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy
import logging

import pytest

from foodx_devops_tools._to import StructuredTo
from foodx_devops_tools.deploy_me.application_steps import deploy_step


@pytest.mark.asyncio
async def test_clean(
    mock_apply_template,
    mock_deploystep_context,
    mock_rg_deploy,
    mock_run_puff,
    mock_verify_puff_target,
):
    await deploy_step(**mock_deploystep_context)

    mock_rg_deploy.assert_called_once()


@pytest.mark.asyncio
async def test_default_override_parameters(
    default_override_parameters,
    mock_apply_template,
    mock_deploystep_context,
    mock_rg_deploy,
    mock_run_puff,
    mock_verify_puff_target,
    mocker,
):
    this_context = copy.deepcopy(mock_deploystep_context)
    this_context["this_step"].static_secrets = False

    await deploy_step(**this_context)

    expected_defaults = default_override_parameters(
        mock_deploystep_context["deployment_data"]
    )
    mock_rg_deploy.assert_called_once_with(
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        deployment_name="app-name_this-step_12345",
        override_parameters=expected_defaults,
        validate=mocker.ANY,
    )


@pytest.mark.asyncio
async def test_secrets_enabled(
    default_override_parameters,
    mock_apply_template,
    mock_deploystep_context,
    mock_rg_deploy,
    mock_run_puff,
    mock_verify_puff_target,
    mocker,
):
    this_context = copy.deepcopy(mock_deploystep_context)
    this_context["this_step"].static_secrets = True

    this_context["deployment_data"].data.static_secrets = {
        "one": 1,
        "two": "2",
    }
    expected_parameters = default_override_parameters(
        mock_deploystep_context["deployment_data"]
    )
    expected_parameters.update(
        {
            "staticSecrets": [
                {
                    "enabled": True,
                    "key": "one",
                    "value": 1,
                },
                {
                    "enabled": True,
                    "key": "two",
                    "value": "2",
                },
            ],
        }
    )

    await deploy_step(**this_context)

    mock_rg_deploy.assert_called_once_with(
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        mocker.ANY,
        deployment_name="app-name_this-step_12345",
        override_parameters=expected_parameters,
        validate=mocker.ANY,
    )


@pytest.mark.asyncio
async def test_step_skip(caplog, mock_deploystep_context, mock_rg_deploy):
    with caplog.at_level(logging.DEBUG):
        mock_deploystep_context["deployment_data"].data.to = StructuredTo(
            frame="f1", application="a1", step="other_step"
        )

        await deploy_step(**mock_deploystep_context)

        mock_rg_deploy.assert_not_called()

        assert (
            "application step skipped using deployment specifier" in caplog.text
        )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import os
import pathlib
import tempfile
import typing

from foodx_devops_tools.utilities.ansible import _encrypt_vault


@contextlib.contextmanager
def encrypted_file(
    content: str, password: str
) -> typing.Generator[typing.Tuple[pathlib.Path, pathlib.Path], None, None]:
    with tempfile.TemporaryDirectory() as dir:
        dir_path = pathlib.Path(dir)
        encrypted_file_path = dir_path / "some.file"

        password_file_path = dir_path / "password-file"
        with password_file_path.open("w") as f:
            f.write(password)

        unencrypted_file_path = dir_path / "plain.file"
        with unencrypted_file_path.open("w") as f:
            f.write(content)

        _encrypt_vault(
            encrypted_file_path, password_file_path, unencrypted_file_path
        )

        os.remove(unencrypted_file_path)
        os.remove(password_file_path)

        assert encrypted_file_path.is_file()
        with encrypted_file_path.open("r") as f:
            encrypted_content = f.read()

        assert encrypted_content.startswith("$ANSIBLE_VAULT;1.1;AES256")

        yield encrypted_file_path



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with foodx_devops_tools.
#  If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.pipeline_config import load_tenants
from foodx_devops_tools.pipeline_config.exceptions import TenantsDefinitionError


@pytest.fixture
def apply_tenants_test(apply_pipeline_config_test):
    def _apply(mock_content: str):
        result = apply_pipeline_config_test(mock_content, load_tenants)
        return result

    return _apply


def test_single(apply_tenants_test):
    file_text = """---
tenants:
  name:
    azure_id: abc123
"""

    result = apply_tenants_test(file_text)

    assert len(result.tenants) == 1
    assert "name" in result.tenants
    assert result.tenants["name"].azure_id == "abc123"
    assert result.tenants["name"].azure_name is None


def test_name(apply_tenants_test):
    file_text = """---
tenants:
  name:
    azure_id: abc123
    azure_name: some name
"""

    result = apply_tenants_test(file_text)

    assert len(result.tenants) == 1
    assert "name" in result.tenants
    assert result.tenants["name"].azure_id == "abc123"
    assert result.tenants["name"].azure_name == "some name"


def test_multiple(apply_tenants_test):
    file_text = """---
tenants:
  name1:
    azure_id: abc123
  name2:
    azure_id: abc1234
  name3:
    azure_id: abc12345
"""

    result = apply_tenants_test(file_text)

    assert len(result.tenants) == 3
    assert all([x in result.tenants for x in ["name1", "name2", "name3"]])
    assert result.tenants["name1"].azure_id == "abc123"


def test_none_raises(apply_tenants_test):
    file_text = """---
"""

    with pytest.raises(
        TenantsDefinitionError,
        match=r"Error validating tenants definition",
    ):
        apply_tenants_test(file_text)


def test_empty_raises(apply_tenants_test):
    file_text = """---
tenants:
"""

    with pytest.raises(
        TenantsDefinitionError,
        match=r"Error validating tenants definition",
    ):
        apply_tenants_test(file_text)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy


def test_clean(mock_flattened_deployment):
    under_test = copy.deepcopy(mock_flattened_deployment[0])
    under_test.context.application_name = "app_name"
    under_test.context.frame_name = "f1"
    under_test.context.pipeline_id = "12345678.24"

    result = under_test.construct_deployment_name("stepname")

    assert result == "app-name_stepname_12345678.24"


def test_plus(mock_flattened_deployment):
    under_test = copy.deepcopy(mock_flattened_deployment[0])
    under_test.context.application_name = "app_name"
    under_test.context.frame_name = "f1"
    under_test.context.pipeline_id = "12345678+24"

    result = under_test.construct_deployment_name("stepname")

    assert result == "app-name_stepname_12345678-24"


def test_underscore(mock_flattened_deployment):
    under_test = copy.deepcopy(mock_flattened_deployment[0])
    under_test.context.application_name = "app_name"
    under_test.context.frame_name = "f1"
    under_test.context.pipeline_id = "12345678_24"

    result = under_test.construct_deployment_name("stepname")

    assert result == "app-name_stepname_12345678-24"



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Manage puff map configuration files."""

import copy
import logging
import pathlib
import typing

import pydantic

from ._exceptions import FolderPathError, PuffMapDefinitionsError
from ._loader import load_configuration
from ._structure import FrameFile, StructuredName, StructuredPathCollection

log = logging.getLogger(__name__)

ENTITY_NAME = "puff_map"
ENTITY_SINGULAR = "puff_map"

PuffMapPaths = typing.Dict[str, pathlib.Path]
PuffMapSubscriptions = typing.Dict[str, PuffMapPaths]
PuffMapReleaseStates = typing.Dict[str, PuffMapSubscriptions]


class PuffMapApplication(pydantic.BaseModel):
    """Singular application collection of puff file mappings."""

    arm_parameters_files: PuffMapReleaseStates


PuffMapApplications = typing.Dict[str, PuffMapApplication]


class PuffMapFrame(pydantic.BaseModel):
    """Singular frame collection of puff file mappings of it's applications."""

    applications: PuffMapApplications


PuffMapFrames = typing.Dict[str, PuffMapFrame]

T = typing.TypeVar("T", bound="PuffMap")


class PuffMap(pydantic.BaseModel):
    """Map puff generated ARM template parameter files to subscriptions."""

    frames: PuffMapFrames

    def arm_template_parameter_file_paths(
        self: T, folder_paths: StructuredPathCollection
    ) -> StructuredPathCollection:
        """
        Generate a collection ARM template parameter file paths from puff_map.

        Args:
            folder_paths: Frame folder paths.

        Returns:
            ARM template parameter file paths indexed by structured name.
        Raises:
            FolderPathError: If there is a problem acquiring frame folder
                             path for the specified frame.
        """
        results: StructuredPathCollection = dict()
        for frame_name, frame_data in self.frames.items():
            frame_structure = StructuredName()
            frame_structure.append(frame_name)
            for app_name, app_data in frame_data.applications.items():
                app_structure = copy.deepcopy(frame_structure)
                app_structure.append(app_name)
                for (
                    release_state,
                    release_data,
                ) in app_data.arm_parameters_files.items():
                    release_structure = copy.deepcopy(app_structure)
                    release_structure.append(release_state)
                    for sub_name, sub_data in release_data.items():
                        sub_structure = copy.deepcopy(release_structure)
                        sub_structure.append(sub_name)
                        for step_name, puff_file in sub_data.items():
                            step_structure = copy.deepcopy(sub_structure)
                            step_structure.append(step_name)

                            try:
                                results[step_structure] = FrameFile(
                                    dir=folder_paths[frame_structure].dir,
                                    file=puff_file,
                                )
                            except KeyError as e:
                                raise FolderPathError(
                                    "Structure for specified folder path does "
                                    "not exist, {0}, {1}".format(
                                        str(frame_structure),
                                        str(step_structure),
                                    )
                                ) from e

        return results


class PuffMapGeneratedFiles(pydantic.BaseModel):
    """Map puff generated ARM template parameter files to subscriptions."""

    puff_map: PuffMap


def load_puff_map(file_path: pathlib.Path) -> PuffMapGeneratedFiles:
    """
    Load puff map definitions from file.

    Args:
        file_path: Path to puff map definitions file.

    Returns:
        Puff map definitions.
    Raises:
        PuffMapDefinitionsError: If an error occurs loading the file.
    """
    log.info("loading frame definitions from file, {0}".format(file_path))
    result = load_configuration(
        file_path, PuffMapGeneratedFiles, PuffMapDefinitionsError, ENTITY_NAME
    )

    return typing.cast(PuffMapGeneratedFiles, result)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import uuid

import pytest

from foodx_devops_tools.puff.arm import (
    _delete_parameter_file,
    _merge_default_name,
    _merge_list_of_dict,
    _remove_keys,
)

from .support import initialize_filesystem


class TestRemoveKeys:
    def test_clean(self):
        data = {
            "name": "some name",
            "p1": "v1",
            "p2": 3.14159,
            "environments": dict(),
            "services": dict(),
        }
        pre_copy = data.copy()
        result = _remove_keys(data, ["name", "environments", "services"])

        assert result == {
            "p1": "v1",
            "p2": 3.14159,
        }
        # original data must not be modified
        assert data == pre_copy


class TestDeleteParameterFile:
    @pytest.mark.asyncio
    async def test_delete_clean(self, tmp_path_factory):
        this_directory = tmp_path_factory.mktemp(str(uuid.uuid4()))
        this_file = this_directory / "some_file"
        initialize_filesystem(this_file, "some content")

        assert this_file.exists()

        await _delete_parameter_file(this_file)

        assert not this_file.exists()


class TestMergeDefaultName:
    def test_default_present(self):
        mock_base = {
            "some-name": {
                "default": {"name": "name-variable-value"},
                "p1": "p1",
            }
        }
        expected_value = {
            "some-name": {"p1": "p1", "name": "name-variable-value"}
        }

        result = _merge_default_name(mock_base)

        assert result == expected_value

    def test_default_absent(self):
        mock_base = {
            "some-name": {
                "p1": "p1",
            }
        }
        expected_value = {
            "some-name": {
                "p1": "p1",
            }
        }

        result = _merge_default_name(mock_base)

        assert result == expected_value


class TestMergeListOfDict:
    def test_clean(self):
        data = [
            {"r1": {"k1": "e1r1k1"}},
            {"r1": {"k2": "e2r1k2"}},
            {"r1": {"k3": "e3r1k3"}},
            {"r2": {"k4": "e1r2k2"}},
            {"r2": {"k5": "e2r2k3"}},
        ]
        expected_value = {
            "r1": {
                "k1": "e1r1k1",
                "k2": "e2r1k2",
                "k3": "e3r1k3",
            },
            "r2": {
                "k4": "e1r2k2",
                "k5": "e2r2k3",
            },
        }

        result = _merge_list_of_dict(data)

        assert result == expected_value

    def test_none_empty_subsequent_entries(self):
        data = [
            {"r1": {"k1": "e1r1k1"}},
            {"r1": None},
            {"r1": {"k3": "e3r1k3"}},
            {"r2": {"k4": "e1r2k2"}},
            {"r2": dict()},
        ]
        expected_value = {
            "r1": {
                "k1": "e1r1k1",
                "k3": "e3r1k3",
            },
            "r2": {
                "k4": "e1r2k2",
            },
        }

        result = _merge_list_of_dict(data)

        assert result == expected_value

    def test_none_empty_initial_entries(self):
        data = [
            {"r1": None},
            {"r1": {"k2": "e2r1k2"}},
            {"r1": {"k3": "e3r1k3"}},
            {"r2": dict()},
            {"r2": {"k5": "e2r2k3"}},
        ]
        expected_value = {
            "r1": {
                "k2": "e2r1k2",
                "k3": "e3r1k3",
            },
            "r2": {
                "k5": "e2r2k3",
            },
        }

        result = _merge_list_of_dict(data)

        assert result == expected_value

    def test_none_empty_only_entries(self):
        data = [
            {"r1": None},
            {"r2": dict()},
        ]
        expected_value = {
            "r1": dict(),
            "r2": dict(),
        }

        result = _merge_list_of_dict(data)

        assert result == expected_value



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Azure resource group utilities."""

import asyncio
import contextlib
import json
import logging
import pathlib
import re
import typing

from foodx_devops_tools.utilities import run_async_command
from foodx_devops_tools.utilities.exceptions import CommandError

from .model import AzureSubscriptionConfiguration

log = logging.getLogger(__name__)

AZURE_GROUP_ID_PATTERN = (
    r"^/subscriptions/"
    r"(?P<subscription_id>[0-9a-z\-]+)"
    r"/resourceGroups/"
    r"(?P<group_name>[A-Za-z0-9\-_]+)"
)
AZURE_GROUP_MATCH = re.compile(AZURE_GROUP_ID_PATTERN)


class ResourceGroupError(Exception):
    """Problem occurred with resource group related actions."""


async def check_exists(
    resource_group_name: str, subscription: AzureSubscriptionConfiguration
) -> dict:
    """
    Determine if a resource group exists in the specified subscription.

    Assume authentication has already occurred such that subsequent ``az``
    CLI commands will succeed.

    Args:
        resource_group_name: Name of resource group to find
        subscription: Subscription to search for resource group

    Returns:
        Query result data on success. Empty dictionary otherwise.
    Raises:
        ResourceGroupError: If any error occurs preventing the search from
                            completing.
    """
    try:
        this_command = [
            "az",
            "group",
            "list",
            "--subscription",
            subscription.subscription_id,
        ]
        log.debug("{0}".format(str(this_command)))
        result = await run_async_command(this_command)
        log.debug(
            "resource group existence check stdout, {0}".format(result.out)
        )
        log.debug(
            "resource group existence check stderr, {0}".format(result.error)
        )

        result_data = json.loads(result.out)
        log.debug("{0}".format(str(result_data)))
        group_result = dict()
        for this_group in result_data:
            this_match = AZURE_GROUP_MATCH.match(this_group["id"])
            if (
                this_match
                and this_match.group("group_name") == resource_group_name
            ):
                group_result = this_group.copy()
                log.debug("group exists, {0}".format(str(group_result)))

        return group_result
    except asyncio.CancelledError:
        # should almost always let async cancelled exceptions propagate.
        raise
    except Exception as e:
        raise ResourceGroupError(
            "Problem acquiring resource group data, {0}, {1}, {2}".format(
                resource_group_name, type(e), str(e)
            )
        ) from e


async def create(
    resource_group_name: str,
    location: str,
    subscription: AzureSubscriptionConfiguration,
) -> None:
    """
    Idempotent creation of an Azure resource group.

    Args:
        resource_group_name: Name of resource group to create.
        location: Azure location in which to create the resource group.
        subscription: Azure subscription identity for the resource group.
    """
    group_data = await check_exists(resource_group_name, subscription)
    if not group_data:
        result = await run_async_command(
            [
                "az",
                "group",
                "create",
                "--resource-group",
                resource_group_name,
                "--location",
                location,
                "--subscription",
                subscription.subscription_id,
            ]
        )
        log.debug("resource group creation stdout, {0}".format(result.out))
        log.debug("resource group creation stderr, {0}".format(result.error))


async def delete(
    resource_group_name: str,
    subscription: AzureSubscriptionConfiguration,
) -> None:
    """
    Delete an Azure resource group.

    Args:
        resource_group_name: Name of resource group to create.
        subscription: Azure subscription identity for the resource group.
    """
    group_data = await check_exists(resource_group_name, subscription)
    if group_data:
        log.info(
            "deleting resource group, {0}, {1}".format(
                resource_group_name, subscription.subscription_id
            )
        )
        result = await run_async_command(
            [
                "az",
                "group",
                "delete",
                "--resource-group",
                resource_group_name,
                "--yes",
                "--subscription",
                subscription.subscription_id,
            ]
        )
        log.debug("resource group deletion stdout, {0}".format(result.out))
        log.debug("resource group deletion stderr, {0}".format(result.error))
    else:
        log.info(
            "no delete resource group as it does not exist, {0}, {1}".format(
                resource_group_name, subscription.subscription_id
            )
        )


def _make_arm_parameter_values(key_values: dict) -> dict:
    """Transform a key-value dictionary into ARM template parameter data."""
    result = dict()
    for k, v in key_values.items():
        result[k] = {"value": v}

    return result


async def deploy(
    resource_group_name: str,
    arm_template_path: pathlib.Path,
    arm_parameters_path: pathlib.Path,
    location: str,
    mode: str,
    subscription: AzureSubscriptionConfiguration,
    deployment_name: typing.Optional[str] = None,
    override_parameters: typing.Optional[dict] = None,
    validate: bool = False,
) -> None:
    """
    Deploy resource to a resource group.

    The resources are defined in ARM template JSON or bicep files, per Azure
    CLI utility.

    Args:
        resource_group_name: Resource group name
        arm_template_path: Path to ARM template deployment file.
        arm_parameters_path: Path to ARM template deployment parameters file.
        location: Azure location in which to deploy resource group.
        mode: Deployment mode; "Complete" or "Incremental".
        subscription: Target subscription/tenant for deployment.
        override_parameters: Key-value pairs of json compatible data to pass
            to deployment. (optional; default None)
        validate: Flag to enable deployment validation. (optional; default
            False)
    """
    result = None
    try:
        await create(resource_group_name, location, subscription)
        # assume az cli is in the PATH when command is run...
        this_command = [
            "az",
            "deployment",
            "group",
            "create" if not validate else "validate",
            "--mode",
            mode,
            "--resource-group",
            resource_group_name,
            "--subscription",
            subscription.subscription_id,
            "--template-file",
            str(arm_template_path),
            "--parameters",
            "@{0}".format(arm_parameters_path),
        ]
        if deployment_name:
            this_command += [
                "--name",
                deployment_name,
            ]
        log.debug("az command, {0}".format(str(this_command)))
        if override_parameters:
            # WARNING: these external parameters may be sensitive content
            # such as secrets, so DO NOT LOG.
            log.debug(
                "az command override parameters specified. content "
                "withheld from log."
            )
            this_command += [
                "--parameters",
                "{0}".format(
                    json.dumps(_make_arm_parameter_values(override_parameters))
                ),
            ]

        result = await run_async_command(this_command)
        log.info(
            "resource group deployment succeeded, {0} ({1})".format(
                resource_group_name, subscription.subscription_id
            )
        )
    except CommandError as e:
        log.error(
            "resource group deployment failed, {0} ({1}), {2}".format(
                resource_group_name, subscription.subscription_id, str(e)
            )
        )
        if result:
            log.debug(
                "resource group deployment stdout, {0} ({1}), {2}".format(
                    resource_group_name,
                    subscription.subscription_id,
                    result.out,
                )
            )
            log.debug(
                "resource group deployment stderr, {0} ({1}), {2}".format(
                    resource_group_name,
                    subscription.subscription_id,
                    result.error,
                )
            )
        else:
            log.error(
                "unexpected error, {0} ({1}), {2}".format(
                    resource_group_name, subscription.subscription_id, str(e)
                )
            )
        raise


@contextlib.asynccontextmanager
async def delete_when_done(
    resource_group_name: str,
    subscription: AzureSubscriptionConfiguration,
) -> typing.AsyncGenerator[None, None]:
    """
    Delete a resource group when this context goes out of scope.

    Should probably not be used for "normal" resource group creation
    deployments, but is likely to be useful for clean up of validation
    deployments.

    Args:
        resource_group_name: Name of resource group to delete.
        subscription: Subscription that "owns" the resource group.
    """
    yield

    # don't block for deletion event.
    await asyncio.shield(
        asyncio.create_task(delete(resource_group_name, subscription))
    )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import re
import subprocess
import typing

MAX_COMMIT_SHA_LENGTH = 7

PACKAGE_EXTENSION = "tgz"

GIT_DESCRIBE_PATTERN = r"^(?P<tag>[0-9a-z\-.]+)-(?P<offset>[0-9]+)-([0-9a-z]+)$"
SEMANTIC_RELEASE_PATTERN = r"(\d+)(\.\d+)*"
SEMANTIC_TAG_PATTERN = (
    r"^(refs/tags/(?P<release>{0}$))|"
    r"(refs/tags/(?P<dryrun>{0}-dryrun\d*)$)".format(SEMANTIC_RELEASE_PATTERN)
)


def acquire_post_data() -> typing.Tuple[str, str]:
    """
    Acquire the offset of HEAD from the last release tag in commit history.

    The offset is calculated as the number of commits between HEAD and the
    release tag.

    Returns:
        Tuple of last release tag and offset.
    """
    run_result = subprocess.run(
        # exclude prerelease tags from git describe
        ["git", "describe", "--first-parent", "--tags", "--exclude", "*.*.*-*"],
        capture_output=True,
        text=True,
    )

    if run_result.returncode != 0:
        raise RuntimeError(
            "git describe command failed, {0}".format(run_result.stderr)
        )

    match_result = re.match(GIT_DESCRIBE_PATTERN, run_result.stdout)
    if not match_result:
        raise RuntimeError(
            "git describe tag error, {0}".format(run_result.stdout)
        )

    return match_result.group("tag"), match_result.group("offset")


def acquire_release_id(ci_ref: str, commit_sha: str) -> str:
    """
    Formulate release id aligned with current CI branch.

    Args:
        ci_ref: CI branch reference.
        commit_sha: CI HEAD git commit SHA.

    Returns:
        Formulated release id.
    """
    if len(commit_sha) > MAX_COMMIT_SHA_LENGTH:
        this_sha = commit_sha[0:MAX_COMMIT_SHA_LENGTH]
    else:
        this_sha = commit_sha

    semantic_commit_ref = re.compile(SEMANTIC_TAG_PATTERN)
    regex_result = semantic_commit_ref.match(ci_ref)
    if regex_result and regex_result.group("release"):
        release_id = regex_result.group("release")
    elif regex_result and regex_result.group("dryrun"):
        release_id = regex_result.group("dryrun")
    else:
        # non-release use "-post" suffix
        this_tag, this_offset = acquire_post_data()
        release_id = "{0}-post.{1}.{2}".format(this_tag, this_offset, this_sha)

    return release_id



# https://gitlab.com/ci-cd-devops/build_harness/-/blob/feature/27-fix-main-test-publish/build_harness/commands/_declarations.py
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.

#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Common constants for command definitions."""

DEFAULT_CONSOLE_LOGGING_ENABLED = False
DEFAULT_FILE_LOGGING_DISABLED = False
DEFAULT_FILE_ROTATION_BACKUPS = 100
DEFAULT_FILE_ROTATION_ENABLED = True
DEFAULT_FILE_ROTATION_SIZE_MB = 10

DEFAULT_LOG_LEVEL = "warning"
VALID_LOG_LEVELS = ["critical", "error", "warning", "info", "debug", "notset"]



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Console related reporting."""

import click


def report_success(message: str) -> None:
    """
    Report a success to console.

    The text "SUCCESS: " is pre-pended to the message for printing.

    Args:
        message: Success message to be reported.
    """
    click.echo(click.style("SUCCESS: {0}".format(message), fg="green"))


def report_failure(message: str) -> None:
    """
    Report a failure to console.

    The text "FAILURE: " is pre-pended to the message for printing.

    Args:
        message: Failure message to be reported.
    """
    click.echo(click.style("FAILURE: {0}".format(message), fg="red"), err=True)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Declarations relating to release management."""

import enum


@enum.unique
class ReleaseState(enum.Enum):
    """Enumeration of release states."""

    ftr = enum.auto()
    dev = enum.auto()
    qa = enum.auto()
    stg = enum.auto()
    prd = enum.auto()



#  Copyright (c) 2022 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.deploy_me.application_steps import script_step
from foodx_devops_tools.deploy_me.application_steps._script import (
    run_async_command,
)


@pytest.fixture()
def mock_run_async_command(mock_async_method):
    this_mock = mock_async_method(
        "foodx_devops_tools.deploy_me.application_steps._script"
        ".run_async_command"
    )
    return this_mock


@pytest.fixture()
def mock_apply_dynamic_template(mocker):
    this_mock = mocker.patch(
        "foodx_devops_tools.deploy_me.application_steps._script"
        ".apply_dynamic_template"
    )
    return this_mock


@pytest.mark.asyncio
async def test_clean(mock_shellstep_context, mock_run_async_command):
    await script_step(**mock_shellstep_context)

    mock_run_async_command.assert_called_once_with(
        [
            "/bin/bash",
            "-c",
            # NOTE: jinja2 strips trailing whitespace by default. this can be
            # disabled but it isn't very important here, so just need to note
            # the apparent discrepancy.
            """do something
make f1""",
        ]
    )


@pytest.mark.asyncio
async def test_bash_script():
    this_command = [
        "/bin/bash",
        "-c",
        """echo 123
echo "o\nn\ne" | grep "o"
""",
    ]
    results = await run_async_command(this_command)

    assert "123" in results.out
    assert "o" in results.out
    assert "n" not in results.out
    assert "e" not in results.out



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Check log file for specific content."""

import logging
import mmap
import pathlib
import typing

log = logging.getLogger(__name__)


class CredentialLeakageError(Exception):
    """Test for credentials in a file has failed."""


def check_credential_leakage(
    secrets: typing.Set[str], log_file: pathlib.Path
) -> None:
    """
    Check a file for the presence of the specified string.

    Args:
        secrets: Text to check for in file.
        log_file: File to check for string.

    Raises:
        CredentialLeakageError: If the specified string is found in the text
                                file.
    """
    tests_failed: typing.List[bool] = list()
    with log_file.open(mode="rb") as f:
        this_map = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
        for this_secret in secrets:
            result = this_map.find(this_secret.encode())
            if result == -1:
                # the string was not found in the file so the test PASSED
                tests_failed.append(False)
            else:
                # the string was found in the file so the test FAILED
                tests_failed.append(True)

    if any(tests_failed):
        log.debug(f"check_credential_leakage, {tests_failed}")
        raise CredentialLeakageError(
            f"Credentials leaked into log file, {log_file}"
        )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import logging
import pathlib
import typing

import aiofiles

from foodx_devops_tools._to import StructuredTo
from foodx_devops_tools.utilities.templates import prepare_deployment_files

from ._structure import StructuredName
from .pipeline import PipelineConfiguration
from .views import DeploymentContext, ReleaseView

log = logging.getLogger(__name__)

FileExistsCollection = typing.Dict[pathlib.Path, bool]


def _collate_results(
    results: typing.List[FileExistsCollection],
) -> FileExistsCollection:
    collated_results: FileExistsCollection = dict()
    for this_result in results:
        collated_results.update(this_result)

    return collated_results


async def _file_exists(file: pathlib.Path) -> bool:
    """Test for file existence asynchronously."""
    try:
        async with aiofiles.open(str(file), mode="r"):
            pass

        exists = True
    except FileNotFoundError:
        exists = False

    return exists


async def _check_paths(
    paths: typing.Iterable[pathlib.Path],
) -> FileExistsCollection:
    results = await asyncio.gather(*[_file_exists(x) for x in paths])

    collection: FileExistsCollection = dict()
    for file, state in zip(paths, results):
        collection[file] = state

    return collection


async def _check_arm_files(
    arm_template_files: typing.List[pathlib.Path],
) -> FileExistsCollection:
    return await _check_paths(arm_template_files)


async def _check_puff_maps(
    arm_parameter_files: typing.List[pathlib.Path],
) -> FileExistsCollection:
    return await _check_paths(arm_parameter_files)


PathList = typing.List[pathlib.Path]


async def _prepare_deployment_files(
    pipeline_configuration: PipelineConfiguration,
) -> typing.Tuple[PathList, PathList]:
    arm_paths = pipeline_configuration.frames.arm_file_paths()
    frame_folders = pipeline_configuration.frames.frame_folders()
    puff_paths = pipeline_configuration.frames.puff_file_paths()
    puff_parameter_paths = (
        pipeline_configuration.puff_map.arm_template_parameter_file_paths(
            pipeline_configuration.frames.frame_folders()
        )
    )

    assert set(puff_paths.keys()) == set(arm_paths.keys())

    templated_arm_files = list()
    expected_arm_parameter_files = list()
    for release_state in pipeline_configuration.release_states:
        base_context = DeploymentContext(
            commit_sha="abc123",
            git_ref=None,
            pipeline_id="000",
            release_id="0.0.0+local",
            release_state=release_state,
        )
        log.info("top-level deployment context, {0}".format(str(base_context)))

        pipeline_state = ReleaseView(pipeline_configuration, base_context)
        deployment_iterations = pipeline_state.flatten(StructuredTo())

        for this_iteration in deployment_iterations:
            for structure_name in arm_paths.keys():
                # have to initialise some data here that in a deployment is
                # fulfilled as part of the deployment.
                this_iteration.context.frame_name = structure_name[0]
                this_iteration.context.application_name = structure_name[1]
                this_iteration.data.frame_folder = frame_folders[
                    StructuredName([this_iteration.context.frame_name])
                ].dir

                puff_map_structure_name = StructuredName(
                    [
                        this_iteration.context.frame_name,
                        this_iteration.context.application_name,
                        release_state,
                        this_iteration.context.azure_subscription_name,
                        structure_name[-1],
                    ]
                )
                template_files = this_iteration.construct_deployment_paths(
                    arm_paths[structure_name].file,
                    puff_paths[structure_name].file,
                    puff_parameter_paths[puff_map_structure_name].file,
                )
                log.debug(
                    f"template files for configuration validation,"
                    f" {puff_map_structure_name},  {template_files}"
                )

                template_parameters = (
                    this_iteration.construct_template_parameters()
                )
                log.debug(
                    f"template parameters applied to configuration "
                    f"validation,"
                    f" {puff_map_structure_name}, {template_parameters}"
                )

                deployment_files = await prepare_deployment_files(
                    template_files,
                    template_parameters,
                )

                templated_arm_files.append(deployment_files.arm_template)
                expected_arm_parameter_files.append(deployment_files.parameters)

    return templated_arm_files, expected_arm_parameter_files


async def do_path_check(pipeline_configuration: PipelineConfiguration) -> None:
    """
    Check that paths in configuration actually exist.

    Check that all the generated ARM template parameter files and ARM
    template files are where they are expected to be for the current
    *client* configuration.
    """
    (
        templated_arm_files,
        expected_arm_parameter_files,
    ) = await _prepare_deployment_files(pipeline_configuration)

    this_futures = await asyncio.gather(
        _check_arm_files(templated_arm_files),
        _check_puff_maps(expected_arm_parameter_files),
        return_exceptions=False,
    )

    done_items = [x for x in this_futures if x]
    collated_results = _collate_results(done_items)
    missing_files = [str(x) for x, y in collated_results.items() if not y]
    if missing_files:
        raise FileNotFoundError(
            "files missing from deployment, {0}".format(str(missing_files))
        )



#  Copyright (c) 2022 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import logging
import re
import typing

import click

from foodx_devops_tools.azure.cloud.resource_group import (
    AzureSubscriptionConfiguration,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    deploy as deploy_resource_group,
)
from foodx_devops_tools.pipeline_config import FlattenedDeployment
from foodx_devops_tools.pipeline_config.frames import (
    ApplicationStepDeploymentDefinition,
)
from foodx_devops_tools.pipeline_config.puff_map import PuffMapPaths
from foodx_devops_tools.utilities.templates import prepare_deployment_files

log = logging.getLogger(__name__)


def _construct_resource_group_name(
    client: str,
    user_resource_group_name: str,
) -> str:
    """Construct a resource group name from deployment context."""
    # want the client id as a prefix for easier sorting & grouping in portal.
    result = "-".join([client, user_resource_group_name])

    return result


def _mangle_validation_resource_group(current_name: str, suffix: str) -> str:
    this_suffix = re.sub(r"[_.+]", "-", suffix)
    mangled_name = f"{current_name}-{this_suffix}"

    return mangled_name


def _make_secrets_object(key_values: dict) -> typing.List[dict]:
    """Construct secrets into object form required by Foodx ARM template."""
    result = list()
    for k, v in key_values.items():
        this_entry = {
            "enabled": True,
            "key": k,
            "value": v,
        }
        result.append(this_entry)

    return result


def _construct_override_parameters(
    deployment_data: FlattenedDeployment,
    static_secrets_enabled: typing.Optional[bool],
    step_context: str,
) -> typing.Dict[str, typing.Any]:
    result: typing.Dict[str, typing.Any] = {
        "locations": {
            "primary": deployment_data.data.location_primary,
            "secondary": deployment_data.data.location_secondary,
        },
        "tags": deployment_data.context.as_dict(),
    }
    if static_secrets_enabled:
        if deployment_data.data.static_secrets:
            # pass static secrets as a single object containing all the
            # secret key-value pairs.
            parameter_object: typing.Dict[str, typing.Any] = {
                "staticSecrets": _make_secrets_object(
                    deployment_data.data.static_secrets
                )
            }
            result.update(parameter_object)
        else:
            log.warning(
                "There are no static_secrets even though secrets have been"
                " enabled, {0}".format(step_context)
            )

    return result


async def _do_step_deployment(
    this_step: ApplicationStepDeploymentDefinition,
    deployment_data: FlattenedDeployment,
    puff_parameter_paths: PuffMapPaths,
    this_context: str,
    enable_validation: bool,
) -> None:
    step_context = f"{this_context}.{this_step.name}"

    log.debug(
        f"deployment_data.context, "
        f"{step_context}, {str(deployment_data.context)}"
    )
    log.debug(
        f"deployment_data.data, {step_context}, "
        f"{str(deployment_data.data.deployment_tuple)}, "
        f"{str(deployment_data.data.location_primary)}, "
        f"{str(deployment_data.data.location_secondary)}, "
        f"{str(deployment_data.data.release_state)}, "
        f"{str(deployment_data.data.iteration_context)}"
    )
    try:
        resource_group = _construct_resource_group_name(
            deployment_data.context.client,
            this_step.resource_group,
        )

        if enable_validation:
            log.info(f"validation deployment enabled, {step_context}")
            resource_group = _mangle_validation_resource_group(
                resource_group,
                deployment_data.context.pipeline_id,
            )
        else:
            log.info(f"deployment enabled, {step_context}")

        template_parameters = deployment_data.construct_template_parameters(
            resource_group
        )
        log.debug(f"template parameters, {step_context}, {template_parameters}")

        template_files = deployment_data.construct_deployment_paths(
            this_step.arm_file,
            this_step.puff_file,
            puff_parameter_paths[this_step.name],
        )
        log.debug(f"template files, {template_files}")

        deployment_files = await prepare_deployment_files(
            template_files,
            template_parameters,
        )

        override_parameters = _construct_override_parameters(
            deployment_data, this_step.static_secrets, step_context
        )
        deployment_name = deployment_data.construct_deployment_name(
            this_step.name
        )
        this_subscription = AzureSubscriptionConfiguration(
            subscription_id=deployment_data.context.azure_subscription_name
        )
        await deploy_resource_group(
            resource_group,
            deployment_files.arm_template,
            deployment_files.parameters,
            deployment_data.data.location_primary,
            this_step.mode.value,
            this_subscription,
            deployment_name=deployment_name,
            override_parameters=override_parameters,
            validate=enable_validation,
        )
    except Exception as e:
        message = f"step deployment failed, {step_context}, {str(e)}"
        log.error(message)
        click.echo(click.style(message, fg="red"), err=True)
        raise


async def deploy_step(
    this_step: ApplicationStepDeploymentDefinition,
    deployment_data: FlattenedDeployment,
    puff_parameter_data: PuffMapPaths,
    this_context: str,
    enable_validation: bool,
) -> None:
    """
    Deploy Azure resources.

    Args:
        this_step: Deployment definition for this step action.
        deployment_data: Deployment context related parameters.
        puff_parameter_data: Puff file parameter data.
        this_context: Structured string context id.
        enable_validation: Enable or disable Azure validation deployment.
    """
    step_context = "{0}.{1}".format(this_context, this_step.name)
    deploy_to = deployment_data.data.to
    if deploy_to.step and (this_step.name != deploy_to.step):
        log.warning(
            "application step skipped using deployment specifier, "
            "{0} skipped {1}".format(str(deploy_to), step_context)
        )
    else:
        await _do_step_deployment(
            this_step,
            deployment_data,
            puff_parameter_data,
            this_context,
            enable_validation,
        )
        log.info("application step succeeded, {0}".format(step_context))



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.deploy_me_entry import flit_entry


class TestFlitEntry:
    def test_clean(self, mocker):
        mock_flow = mocker.patch("foodx_devops_tools.deploy_me_entry.deploy_me")

        flit_entry()

        mock_flow.assert_called_once_with()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Generalized structured name for deep referencing objects in data model."""

import dataclasses
import pathlib
import typing

U = typing.TypeVar("U", bound="FrameFile")


@dataclasses.dataclass
class FrameFile:
    """
    Record directory and file paths of a frame separately.

    In a frame the frame folder must eventually be combined with a relative
    file path to access a file, but preprocessing often requires the two to
    be separate, but conceptually linked until needed.
    """

    dir: typing.Optional[pathlib.Path]
    file: typing.Optional[pathlib.Path]

    @property
    def path(self: U) -> typing.Optional[pathlib.Path]:
        """Combine the directory and file paths."""
        result = None
        if self.dir and self.file:
            result = self.dir / self.file
        elif self.dir:
            result = self.dir
        elif self.file:
            result = self.file

        return result


T = typing.TypeVar("T", bound="StructuredName")


class StructuredName(typing.List[str]):
    """Structured name of frames, applications and application steps."""

    def __hash__(self: T) -> int:  # type: ignore[override]
        """Hash ``StructuredName`` object for use as dictionary keys."""
        return hash(str(self))

    def __str__(self: T) -> str:
        """Convert ``StructuredName`` object to a string representation."""
        return ".".join(self)


StructuredPathCollection = typing.Dict[StructuredName, FrameFile]



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import pytest

from foodx_devops_tools.pipeline_config import StructuredName, load_puff_map
from foodx_devops_tools.pipeline_config._structure import FrameFile


@pytest.fixture
def apply_puffmap_test(apply_pipeline_config_test):
    def _apply(mock_content: str):
        result = apply_pipeline_config_test(mock_content, load_puff_map)
        return result

    return _apply


def test_single_default(apply_puffmap_test):
    file_text = """---
puff_map:
  frames:
      f1:
        applications:
          a1:
            arm_parameters_files:
              r1:
                sub1: 
                  label1: some/path/puff1.json
"""

    result = apply_puffmap_test(file_text)

    result_map = result.puff_map
    assert len(result_map.frames) == 1
    assert len(result_map.frames["f1"].applications) == 1
    assert (
        len(result_map.frames["f1"].applications["a1"].arm_parameters_files)
        == 1
    )
    assert (
        len(
            result_map.frames["f1"]
            .applications["a1"]
            .arm_parameters_files["r1"]
        )
        == 1
    )
    assert (
        len(
            result_map.frames["f1"]
            .applications["a1"]
            .arm_parameters_files["r1"]["sub1"]
        )
        == 1
    )
    assert result_map.frames["f1"].applications["a1"].arm_parameters_files[
        "r1"
    ]["sub1"]["label1"] == pathlib.Path("some/path/puff1.json")


class TestPuffFiles:
    mock_folders = {
        StructuredName(["f1"]): FrameFile(
            dir=pathlib.Path("this/dir"), file=pathlib.Path("")
        ),
        StructuredName(["f2"]): FrameFile(
            dir=pathlib.Path("other/dir"), file=pathlib.Path("")
        ),
    }

    def test_single(self, apply_puffmap_test):
        file_text = """---
    puff_map:
      frames:
          f1:
            applications:
              a1:
                arm_parameters_files:
                  r1:
                    sub1: 
                      label1: some/puff1.json
    """

        under_test = apply_puffmap_test(file_text)
        result = under_test.puff_map.arm_template_parameter_file_paths(
            self.mock_folders
        )

        assert result == {
            StructuredName(["f1", "a1", "r1", "sub1", "label1"]): FrameFile(
                dir=pathlib.Path("this/dir/"),
                file=pathlib.Path("some/puff1.json"),
            ),
        }

    def test_multiple_steps(self, apply_puffmap_test):
        file_text = """---
    puff_map:
      frames:
          f1:
            applications:
              a1:
                arm_parameters_files:
                  r1:
                    sub1: 
                      label1: puff1.json
                      label2: puff2.json
    """

        under_test = apply_puffmap_test(file_text)
        result = under_test.puff_map.arm_template_parameter_file_paths(
            self.mock_folders
        )

        assert result == {
            StructuredName(["f1", "a1", "r1", "sub1", "label1"]): FrameFile(
                dir=pathlib.Path("this/dir/"), file=pathlib.Path("puff1.json")
            ),
            StructuredName(["f1", "a1", "r1", "sub1", "label2"]): FrameFile(
                dir=pathlib.Path("this/dir/"), file=pathlib.Path("puff2.json")
            ),
        }

    def test_multiple_frames(self, apply_puffmap_test):
        file_text = """---
    puff_map:
      frames:
          f1:
            applications:
              a1:
                arm_parameters_files:
                  r1:
                    sub1: 
                      label1: some/path/puff1.json
          f2:
            applications:
              a1:
                arm_parameters_files:
                  r1:
                    sub1: 
                      label1: some/path/puff1.json
    """

        under_test = apply_puffmap_test(file_text)
        result = under_test.puff_map.arm_template_parameter_file_paths(
            self.mock_folders
        )

        assert result == {
            StructuredName(["f1", "a1", "r1", "sub1", "label1"]): FrameFile(
                dir=pathlib.Path("this/dir/"),
                file=pathlib.Path("some/path/puff1.json"),
            ),
            StructuredName(["f2", "a1", "r1", "sub1", "label1"]): FrameFile(
                dir=pathlib.Path("other/dir/"),
                file=pathlib.Path("some/path/puff1.json"),
            ),
        }

    def test_multiple_applications(self, apply_puffmap_test):
        file_text = """---
    puff_map:
      frames:
          f1:
            applications:
              a1:
                arm_parameters_files:
                  r1:
                    sub1: 
                      label1: path/puff1.json
              a2:
                arm_parameters_files:
                  r1:
                    sub1: 
                      label1: puff2.json
    """

        under_test = apply_puffmap_test(file_text)
        result = under_test.puff_map.arm_template_parameter_file_paths(
            self.mock_folders
        )

        assert result == {
            StructuredName(["f1", "a1", "r1", "sub1", "label1"]): FrameFile(
                dir=pathlib.Path("this/dir/"),
                file=pathlib.Path("path/puff1.json"),
            ),
            StructuredName(["f1", "a2", "r1", "sub1", "label1"]): FrameFile(
                dir=pathlib.Path("this/dir/"), file=pathlib.Path("puff2.json")
            ),
        }

    def test_mixed(self, apply_puffmap_test):
        file_text = """---
    puff_map:
      frames:
          f1:
            applications:
              a1:
                arm_parameters_files:
                  r1:
                    sub1: 
                      label1: puff1.json
              a2:
                arm_parameters_files:
                  r1:
                    sub1: 
                      label2: puff2.json
                    sub2: 
                      label3: puff3.json
                  r2:
                    sub1: 
                      label4: puff4.json
          f2:
            applications:
              a1:
                arm_parameters_files:
                  r1:
                    sub1: 
                      label1: puff5.json
    """

        under_test = apply_puffmap_test(file_text)
        result = under_test.puff_map.arm_template_parameter_file_paths(
            self.mock_folders
        )

        assert result == {
            StructuredName(["f1", "a1", "r1", "sub1", "label1"]): FrameFile(
                dir=pathlib.Path("this/dir/"), file=pathlib.Path("puff1.json")
            ),
            StructuredName(["f1", "a2", "r1", "sub1", "label2"]): FrameFile(
                dir=pathlib.Path("this/dir/"), file=pathlib.Path("puff2.json")
            ),
            StructuredName(["f1", "a2", "r1", "sub2", "label3"]): FrameFile(
                dir=pathlib.Path("this/dir/"), file=pathlib.Path("puff3.json")
            ),
            StructuredName(["f1", "a2", "r2", "sub1", "label4"]): FrameFile(
                dir=pathlib.Path("this/dir/"), file=pathlib.Path("puff4.json")
            ),
            StructuredName(["f2", "a1", "r1", "sub1", "label1"]): FrameFile(
                dir=pathlib.Path("other/dir/"), file=pathlib.Path("puff5.json")
            ),
        }



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib
import tempfile

import pytest

from foodx_devops_tools._log_check import (
    CredentialLeakageError,
    check_credential_leakage,
)

EXAMPLE_CONTENT = """
some
really{0}
long
content
"""


def test_detected_secret():
    expected_string = "thisismysecret"

    with tempfile.TemporaryDirectory() as dir:
        file_path = pathlib.Path(dir) / "sample"
        with (file_path).open(mode="w") as f:
            # ensure the secret is in the file so the check can fail
            f.write(EXAMPLE_CONTENT.format(expected_string))

        with pytest.raises(
            CredentialLeakageError, match=r"^Credentials leaked into log file"
        ):
            check_credential_leakage({expected_string}, file_path)


def test_detected_multiple_secrets():
    expected_secrets = {"notinfile", "long"}

    with tempfile.TemporaryDirectory() as dir:
        file_path = pathlib.Path(dir) / "sample"
        with (file_path).open(mode="w") as f:
            # ensure the secret is in the file so the check can fail
            f.write(EXAMPLE_CONTENT.format(expected_secrets))

        with pytest.raises(
            CredentialLeakageError, match=r"^Credentials leaked into log file"
        ):
            check_credential_leakage(expected_secrets, file_path)


def test_not_detected_secret():
    expected_string = "thisismysecret"

    with tempfile.TemporaryDirectory() as dir:
        file_path = pathlib.Path(dir) / "sample"
        with (file_path).open(mode="w") as f:
            # ensure the secret is not in the file so the check runs clean
            f.write(EXAMPLE_CONTENT.format(""))

        check_credential_leakage({expected_string}, file_path)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.


import pytest

from foodx_devops_tools.utilities import get_sha


@pytest.fixture()
def mock_gitsha_result(mocker):
    def _apply(returncode: int, output: str = ""):
        mock_result = mocker.MagicMock()
        mock_result.returncode = returncode
        mock_result.stdout = output
        this_mock = mocker.patch(
            "foodx_devops_tools.utilities.git.run_command",
            return_value=mock_result,
        )
        return this_mock

    return _apply


class TestGetSha:
    def test_clean(self, mock_gitsha_result):
        mock_gitsha_result(0, output="abc123")
        result = get_sha()

        assert result == "abc123"

    def test_error_raises(self, mock_gitsha_result):
        mock_gitsha_result(1)
        with pytest.raises(RuntimeError):
            get_sha()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Context sensitive views of pipeline configuration data."""

import copy
import dataclasses
import logging
import pathlib
import re
import typing

import deepmerge  # type: ignore

from foodx_devops_tools._to import StructuredTo
from foodx_devops_tools.azure.cloud import AzureCredentials
from foodx_devops_tools.patterns import SubscriptionData
from foodx_devops_tools.utilities.jinja2 import TemplateParameters
from foodx_devops_tools.utilities.templates import TemplateFiles, TemplatePaths

from ..deployment import DeploymentTuple
from ._exceptions import PipelineViewError
from .pipeline import PipelineConfiguration, PuffMap

log = logging.getLogger(__name__)

Z = typing.TypeVar("Z", bound="IterationContext")


class IterationContext(typing.List[str]):
    """Record the iteration path."""

    def __str__(self: Z) -> str:
        """Convert the iteration path to a structured string."""
        return ".".join(self)


Y = typing.TypeVar("Y", bound="DeploymentContext")


class DeploymentContext:
    """Deployment context data expected to be applied to resource tags."""

    commit_sha: str
    git_ref: typing.Optional[str]
    pipeline_id: str
    release_id: str
    release_state: str

    # these data must be provided at deployment time because they are more
    # deeply context sensitive.
    __application_name: typing.Optional[str] = None
    __azure_subscription_name: typing.Optional[str] = None
    __azure_tenant_name: typing.Optional[str] = None
    __client: typing.Optional[str] = None
    __frame_name: typing.Optional[str] = None
    __system: typing.Optional[str] = None

    def __init__(
        self: Y,
        commit_sha: str,
        git_ref: typing.Optional[str],
        pipeline_id: str,
        release_id: str,
        release_state: str,
    ) -> None:
        """Construct ``DeploymentContext`` object."""
        self.commit_sha = commit_sha
        self.git_ref = git_ref
        self.pipeline_id = pipeline_id
        self.release_id = release_id
        self.release_state = release_state

    @property
    def application_name(self: Y) -> str:
        """Get application name property."""
        if self.__application_name:
            return self.__application_name
        else:
            raise PipelineViewError(
                "application name not specified in deployment data"
            )

    @application_name.setter
    def application_name(self: Y, v: str) -> None:
        """Set application name property."""
        self.__application_name = v

    @property
    def azure_subscription_name(self: Y) -> str:
        """Get azure subscription name property."""
        if self.__azure_subscription_name:
            return self.__azure_subscription_name
        else:
            raise PipelineViewError(
                "azure subscription name not specified in deployment data"
            )

    @azure_subscription_name.setter
    def azure_subscription_name(self: Y, v: str) -> None:
        """Set azure subscription name property."""
        self.__azure_subscription_name = v

    @property
    def azure_tenant_name(self: Y) -> str:
        """Get azure tenant name property."""
        if self.__azure_tenant_name:
            return self.__azure_tenant_name
        else:
            raise PipelineViewError(
                "azure tenant name not specified in deployment data"
            )

    @azure_tenant_name.setter
    def azure_tenant_name(self: Y, v: str) -> None:
        """Set azure tenant name property."""
        self.__azure_tenant_name = v

    @property
    def client(self: Y) -> str:
        """Get client name property."""
        if self.__client:
            return self.__client
        else:
            raise PipelineViewError("client not specified in deployment data")

    @client.setter
    def client(self: Y, v: str) -> None:
        """Set client name property."""
        self.__client = v

    @property
    def frame_name(self: Y) -> str:
        """Get frame name property."""
        if self.__frame_name:
            return self.__frame_name
        else:
            raise PipelineViewError(
                "frame name not specified in deployment data"
            )

    @frame_name.setter
    def frame_name(self: Y, v: str) -> None:
        """Set frame name property."""
        self.__frame_name = v

    @property
    def resource_suffix(self: Y) -> str:
        """Get resource suffix property."""
        try:
            subscription_data = SubscriptionData.from_subscription_name(
                self.azure_subscription_name
            )

            return subscription_data.resource_suffix
        except PipelineViewError:
            # azure_subscription_name is undefined so just return an
            # identifiable response here.
            return "UNSPECIFIED"

    @property
    def system(self: Y) -> str:
        """Get system name property."""
        if self.__system:
            return self.__system
        else:
            raise PipelineViewError("system not specified in deployment data")

    @system.setter
    def system(self: Y, v: str) -> None:
        """Set system name property."""
        self.__system = v

    def as_dict(self: Y) -> dict:
        """
        Generate ``dict`` object representation of context data.

        The representation is intended to be used for tags data in ARM
        template parameter files. Note that "microsoft", "azure" and
        "windows" are reserved prefixes in Azure tags so any fields using
        these prefixes must be renamed for the dictionary conversion.
        """
        return {
            "application_name": self.__application_name,
            "client": self.__client,
            "commit_sha": self.commit_sha,
            "frame_name": self.__frame_name,
            "pipeline_id": self.pipeline_id,
            "release_id": self.release_id,
            "release_state": self.release_state,
            "resource_suffix": self.resource_suffix,
            "subscription_name": self.__azure_subscription_name,
            "system": self.__system,
            "tenant_name": self.__azure_tenant_name,
        }

    def __str__(self: Y) -> str:
        """Convert object to str for logging purposes."""
        return str(self.as_dict())


X = typing.TypeVar("X", bound="DeployDataView")


class DeployDataView:
    """Data critical to a resource deployment."""

    azure_credentials: AzureCredentials
    deployment_tuple: str
    location_primary: str
    release_state: str
    root_fqdn: str
    static_secrets: dict
    subscription_id: str
    template_context: dict
    tenant_id: str
    url_endpoints: typing.List[str]

    frame_folder: typing.Optional[pathlib.Path] = None

    __location_secondary: typing.Optional[str] = None

    iteration_context: IterationContext
    to: StructuredTo
    __puff_map: typing.Optional[PuffMap] = None

    def __init__(
        self: X,
        azure_credentials: AzureCredentials,
        deployment_tuple: str,
        location_primary: str,
        release_state: str,
        root_fqdn: str,
        static_secrets: dict,
        subscription_id: str,
        tenant_id: str,
        url_endpoints: typing.List[str],
        user_defined_template_context: dict,
        location_secondary: typing.Optional[str] = None,
    ) -> None:
        """Construct ``DeployDataView`` object."""
        self.azure_credentials = azure_credentials
        self.deployment_tuple = deployment_tuple
        self.location_primary = location_primary
        self.release_state = release_state
        self.root_fqdn = root_fqdn
        self.static_secrets = static_secrets
        self.subscription_id = subscription_id
        self.tenant_id = tenant_id
        self.template_context = user_defined_template_context
        self.url_endpoints = url_endpoints
        self.__location_secondary = location_secondary

        self.iteration_context = IterationContext()
        self.to = StructuredTo()

    @property
    def location_secondary(self: X) -> typing.Optional[str]:
        """Get location secondary property."""
        return self.__location_secondary

    @location_secondary.setter
    def location_secondary(self: X, v: str) -> None:
        """Set location secondary property."""
        self.__location_secondary = v

    @property
    def puff_map(self: X) -> PuffMap:
        """Get puff map property."""
        if self.__puff_map:
            return self.__puff_map
        else:
            raise PipelineViewError("puff map not specified in deployment data")

    @puff_map.setter
    def puff_map(self: X, value: PuffMap) -> None:
        """Set puff map property."""
        self.__puff_map = value

    def __str__(self: X) -> str:
        """Convert object to str for logging purposes."""
        return str(
            {
                "azure_credentials": self.azure_credentials,
                "deployment_tuple": self.deployment_tuple,
                "location_primary": self.location_primary,
                "release_state": self.release_state,
                "location_secondary": self.__location_secondary,
                "iteration_context": self.iteration_context,
            }
        )


W = typing.TypeVar("W", bound="FlattenedDeployment")


@dataclasses.dataclass
class FlattenedDeployment:
    """Flattened deployment data for deployment concurrency."""

    context: DeploymentContext
    data: DeployDataView

    def copy_add_frame(self: W, frame_name: str) -> W:
        """
        Deep copy this object and add frame name data.

        Args:
            frame_name: Frame name data to add.

        Returns:
            Copied and updated object.
        """
        x = copy.deepcopy(self)
        x.context.frame_name = frame_name
        x.data.iteration_context.append(frame_name)
        return x

    def copy_add_application(self: W, application_name: str) -> W:
        """
        Deep copy this object and add application name data.

        Args:
            application_name: Application name data to add.

        Returns:
            Copied and updated object.
        """
        x = copy.deepcopy(self)
        x.context.application_name = application_name
        x.data.iteration_context.append(application_name)
        return x

    def construct_fqdn(self: W, leaf_name: str) -> str:
        """
        Construct an FQDN from deployment context.

        Raises:
            SubscriptionNameError: If the subscription name cannot be parsed to
                extract the resource suffix.
        """
        subscription_data = SubscriptionData.from_subscription_name(
            self.context.azure_subscription_name
        )

        return ".".join(
            [
                leaf_name,
                subscription_data.resource_suffix,
                self.context.client,
                self.data.root_fqdn,
            ]
        )

    def construct_app_fqdns(self: W) -> TemplateParameters:
        """Construct endpoint FQDNs for deployyment."""
        result: TemplateParameters = {
            x: self.construct_fqdn(x) for x in self.data.url_endpoints
        }
        result["root"] = self.data.root_fqdn
        result["support"] = "support.{0}.{1}".format(
            self.context.client, self.data.root_fqdn
        )
        return result

    def construct_app_urls(self: W) -> TemplateParameters:
        """Construct endpoint URLs for deployment."""
        fqdns = self.construct_app_fqdns()
        result: TemplateParameters = {
            x: "https://{0}".format(y) for x, y in fqdns.items() if x != "root"
        }

        return result

    def construct_template_parameters(
        self: W, resource_group_name: typing.Optional[str] = None
    ) -> TemplateParameters:
        """
        Construct set of parameters for jinja2 templates.

        Args:
            resource_group_name:    Name of current resource group being
                                    deployed to.

        Returns:
            Dict of parameters to be applied to jinja2 templating.
        """
        engine_data = {
            "environment": {
                "azure": {
                    "subscription_id": self.data.subscription_id,
                    "tenant_id": self.data.tenant_id,
                },
                "resource_group": resource_group_name,
            },
            "locations": {
                "primary": self.data.location_primary,
                "secondary": self.data.location_secondary,
            },
            "network": {
                "fqdns": self.construct_app_fqdns(),
                "urls": self.construct_app_urls(),
            },
            "tags": self.context.as_dict(),
        }
        result: TemplateParameters = {
            "context": {
                **deepmerge.always_merger.merge(
                    copy.deepcopy(self.data.template_context), engine_data
                ),
            },
        }

        return result

    def construct_deployment_name(self: W, step_name: str) -> str:
        """
        Construct the deployment name for use in az CLI.

        Note Azure deployment name have specific limitations:

        * limited to 64 characters
        * must only contain alphanumerics and the characters ".-_"
        """
        assert self.context.application_name is not None
        assert self.context.pipeline_id is not None
        assert self.context.release_id is not None

        substitution_value = "-"
        # reserving underscore here for segmentation of deployment name.
        regex = re.compile(r"[^A-Za-z0-9.\-]")

        filtered_app_name = regex.sub(
            substitution_value, self.context.application_name
        )[0:20]
        filtered_pipeline_id = regex.sub(
            substitution_value, self.context.pipeline_id
        )
        filtered_step_name = regex.sub(substitution_value, step_name)[0:20]

        result = "{0}_{1}".format(
            filtered_app_name
            if filtered_app_name == filtered_step_name
            else "{0}_{1}".format(filtered_app_name, filtered_step_name),
            filtered_pipeline_id,
        )

        return result[0:64]

    @staticmethod
    def __construct_working_directory(
        parent_dir: pathlib.Path, working_name: str
    ) -> pathlib.Path:
        working_dir = parent_dir / "working" / working_name

        return working_dir

    @staticmethod
    def __encode_working_name(parameters: TemplateParameters) -> str:
        """Construct a working directory name from template parameters."""
        # for now, use a simple predictable, non-empty value
        value = "w"
        return value

    def construct_deployment_paths(
        self: W,
        specified_arm_file: typing.Optional[pathlib.Path],
        specified_puff_file: typing.Optional[pathlib.Path],
        target_arm_parameter_path: typing.Optional[pathlib.Path],
    ) -> TemplateFiles:
        """
        Construct paths for ARM template files.

        Args:
            specified_arm_file:
            specified_puff_file:
            target_arm_parameter_path:
            working_name:

        Returns:
            Tuple of necessary paths.
        Raises:
            PipelineViewError:  If any errors occur due to undefined deployment
                                data.
        """
        template_parameters = self.construct_template_parameters()
        working_name = self.__encode_working_name(template_parameters)
        application_name = self.context.application_name
        frame_folder = self.data.frame_folder
        if not frame_folder:
            raise PipelineViewError(
                "frame_folder not defined for application step"
            )
        if not target_arm_parameter_path:
            raise PipelineViewError(
                "target_arm_parameter_path not defined for application step"
            )

        source_arm_template_path = (
            (frame_folder / specified_arm_file)
            if specified_arm_file
            else (frame_folder / "{0}.json".format(application_name))
        )
        log.debug(f"source arm template path, {source_arm_template_path}")

        # Assume that the source arm template path can be used for the other
        # files. In the case that the ARM template path is "out of frame"
        # then the user must explicitly define the puff file path
        if specified_puff_file:
            puff_prompt = "user specified puff file"
            source_puff_path = frame_folder / specified_puff_file
        elif specified_arm_file:
            puff_prompt = "puff file inferred from arm file"
            source_puff_path = source_arm_template_path.parent / str(
                specified_arm_file.name
            ).replace("json", "yml")
        else:
            puff_prompt = "puff file inferred from application name"
            source_puff_path = (
                source_arm_template_path.parent
                / "{0}.yml".format(application_name)
            )
        log.debug(f"{puff_prompt}, {source_puff_path}")

        working_dir = self.__construct_working_directory(
            frame_folder, working_name
        )
        # Assume arm template parameters file has been specified with any sub
        # directories in it's path in puff_map.yml, so only frame_folder is
        # used here.
        parameters_path = working_dir / target_arm_parameter_path
        log.debug(f"arm parameters path, {parameters_path}")

        log.debug(f"working directory, {working_dir}")
        target_arm_template_path = self.__screen_jinja_template(
            source_arm_template_path, working_dir
        )

        template_files = TemplateFiles(
            arm_template=TemplatePaths(
                source=source_arm_template_path,
                target=target_arm_template_path,
            ),
            arm_template_parameters=TemplatePaths(
                source=source_puff_path,
                target=parameters_path,
            ),
        )

        return template_files

    @staticmethod
    def __screen_jinja_template(
        source_path: pathlib.Path, working_dir: pathlib.Path
    ) -> pathlib.Path:
        # A jinja template file needs to have a target in the working dir.
        target_path = working_dir / "{0}".format(source_path.name)

        return target_path


V = typing.TypeVar("V", bound="SubscriptionView")
U = typing.TypeVar("U", bound="DeploymentView")
T = typing.TypeVar("T", bound="ReleaseView")


class SubscriptionView:
    """View of pipeline configuration conditioned by a subscription name."""

    deployment_view: "DeploymentView"
    subscription_name: str

    def __init__(self: V, deployment_view: U, subscription_name: str) -> None:
        """Construct ``SubscriptionView`` object."""
        self.deployment_view = deployment_view
        self.subscription_name = subscription_name

        self._validate_subscription()

    @property
    def deploy_data(self: V) -> typing.List[DeployDataView]:
        """Provide deployment data for each location in this subscription."""
        base_data = (
            self.deployment_view.release_view.configuration.subscriptions[
                self.subscription_name
            ]
        )
        result: typing.List[DeployDataView] = list()
        this_deployment = self.deployment_view.release_view.configuration.deployments.deployment_tuples[  # noqa: E501
            str(self.deployment_view.deployment_tuple)
        ]
        this_service_principals = (
            self.deployment_view.release_view.configuration.service_principals  # noqa: E501
        )
        if not this_service_principals:
            raise PipelineViewError("Missing service principal credentials")

        this_tenants = self.deployment_view.release_view.configuration.tenants
        for this_locations in this_deployment.subscriptions[
            self.subscription_name
        ].locations:
            this_credentials = AzureCredentials(
                userid=this_service_principals[self.subscription_name].id,
                secret=this_service_principals[self.subscription_name].secret,
                name=this_service_principals[self.subscription_name].name,
                subscription=base_data.azure_id,
                tenant=this_tenants[base_data.tenant].azure_id,
            )
            static_secrets = dict()
            secrets_collection = (
                self.deployment_view.release_view.configuration.static_secrets
            )
            if secrets_collection and (
                self.subscription_name in secrets_collection
            ):
                log.debug(
                    "static secret keys, {0}".format(secrets_collection.keys())
                )
                static_secrets = secrets_collection[self.subscription_name]
            elif secrets_collection and (
                self.subscription_name not in secrets_collection
            ):
                raise PipelineViewError(
                    "subscription not defined in static "
                    "secrets, {0}".format(self.subscription_name)
                )

            subscription_id = (
                self.deployment_view.release_view.configuration.subscriptions[
                    self.subscription_name
                ].azure_id
            )
            tenant_name = (
                self.deployment_view.release_view.configuration.subscriptions[
                    self.subscription_name
                ].tenant
            )
            tenant_id = self.deployment_view.release_view.configuration.tenants[
                tenant_name
            ].azure_id
            this_data: typing.Dict[str, typing.Any] = {
                "azure_credentials": this_credentials,
                "deployment_tuple": str(self.deployment_view.deployment_tuple),
                "location_primary": this_locations.primary,
                "location_secondary": this_locations.secondary,
                "root_fqdn": this_deployment.subscriptions[
                    self.subscription_name
                ].root_fqdn,
                "release_state": self.deployment_view.release_view.deployment_context.release_state,  # noqa: E501
                "static_secrets": static_secrets,
                "subscription_id": subscription_id,
                "tenant_id": tenant_id,
                "url_endpoints": self.deployment_view.release_view.configuration.deployments.url_endpoints,  # noqa: E501
                "user_defined_template_context": self.deployment_view.release_view.configuration.context,  # noqa: E501
            }
            if this_locations.secondary:
                this_data["location_secondary"] = this_locations.secondary

            result.append(DeployDataView(**this_data))

        return result

    def _validate_subscription(self: V) -> None:
        if (
            self.subscription_name
            not in self.deployment_view.release_view.configuration.subscriptions
        ):
            raise PipelineViewError(
                "Bad subscription, {0}".format(str(self.subscription_name))
            )


class DeploymentView:
    """View of pipeline configuration conditioned by deployment state."""

    release_view: "ReleaseView"
    deployment_tuple: DeploymentTuple

    def __init__(
        self: U, release_view: T, deployment_state: DeploymentTuple
    ) -> None:
        """Construct ``DeploymentView`` object."""
        self.release_view = release_view
        self.deployment_tuple = deployment_state

        self._validate_deployment_tuple()

    def __matched_subscription_patterns(
        self: U,
        subscription_name: str,
        gitref_patterns: typing.Optional[typing.List[str]],
    ) -> bool:
        """
        Check that the subscription matches any specified patterns.

        If a git reference has not been specified then no conditioning is
        applied and any specified subscriptions will be considered a "match".
        """
        result = True
        this_ref = self.release_view.deployment_context.git_ref
        if this_ref and gitref_patterns:
            not_match = [
                (re.match(x, this_ref) is None) for x in gitref_patterns
            ]
            if all(not_match):
                result = False

        return result

    @property
    def subscriptions(self: U) -> typing.List[SubscriptionView]:
        """
        Provide the subscriptions in this deployment.

        Qualifies subscriptions both by their presence in the deployment
        definitions for the project, and the optional branch patterns
        specified for the subscription.

        Returns:
            List of subscription views in this deployment.
        """
        result: typing.List[SubscriptionView] = list()
        this_id = str(self.deployment_tuple)
        deployment_ids = (
            self.release_view.configuration.deployments.deployment_tuples
        )
        if this_id in deployment_ids:
            for subscription_name in deployment_ids[
                this_id
            ].subscriptions.keys():
                this_subscription = deployment_ids[this_id].subscriptions[
                    subscription_name
                ]
                gitref_patterns = this_subscription.gitref_patterns
                if self.__matched_subscription_patterns(
                    subscription_name, gitref_patterns
                ):
                    result.append(SubscriptionView(self, subscription_name))

        return result

    def _validate_deployment_tuple(self: U) -> None:
        if (
            self.deployment_tuple.release_state
            not in self.release_view.configuration.release_states
        ):
            raise PipelineViewError(
                "Bad release state, {0}".format(str(self.deployment_tuple))
            )

        if (
            self.deployment_tuple.client
            not in self.release_view.configuration.clients.keys()
        ):
            raise PipelineViewError(
                "Bad client, {0}".format(str(self.deployment_tuple))
            )

        if (
            self.deployment_tuple.system
            not in self.release_view.configuration.systems
        ):
            raise PipelineViewError(
                "Bad system, {0}".format(str(self.deployment_tuple))
            )


class ReleaseView:
    """View of pipeline configuration conditioned by release state."""

    configuration: PipelineConfiguration
    deployment_context: DeploymentContext

    def __init__(
        self: T,
        configuration: PipelineConfiguration,
        deployment_context: DeploymentContext,
    ) -> None:
        """Construct ``ReleaseView`` object."""
        self.configuration = configuration
        self.deployment_context = deployment_context

        self._validate_release_state()

    @property
    def deployments(self: T) -> typing.List[DeploymentView]:
        """Provide the deployments in this release."""
        result: typing.List[DeploymentView] = list()
        for this_client in self.configuration.clients.keys():
            for this_system in self.configuration.systems:
                this_state = DeploymentTuple(
                    client=this_client,
                    release_state=self.deployment_context.release_state,
                    system=this_system,
                )
                if (
                    str(this_state)
                    in self.configuration.deployments.deployment_tuples.keys()
                ):
                    result.append(DeploymentView(self, this_state))

        return result

    def flatten(
        self: T,
        to: StructuredTo,
    ) -> typing.List[FlattenedDeployment]:
        """Flatten the nested hierarchy of views into a simple list."""
        result: typing.List[FlattenedDeployment] = list()
        for this_deployment in self.deployments:
            for this_subscription in this_deployment.subscriptions:
                for this_deploy_data in this_subscription.deploy_data:
                    updated_context = copy.deepcopy(self.deployment_context)
                    updated_context.azure_subscription_name = (
                        this_subscription.subscription_name
                    )
                    updated_context.azure_tenant_name = (
                        self.configuration.subscriptions[
                            this_subscription.subscription_name
                        ].tenant
                    )
                    updated_context.client = (
                        this_deployment.deployment_tuple.client
                    )
                    updated_context.system = (
                        this_deployment.deployment_tuple.system
                    )

                    updated_data = copy.deepcopy(this_deploy_data)
                    updated_data.to = to

                    this_value = FlattenedDeployment(
                        context=updated_context,
                        data=updated_data,
                    )
                    result.append(this_value)
        return result

    def _validate_release_state(self: T) -> None:
        if (
            self.deployment_context.release_state
            not in self.configuration.release_states
        ):
            message = "Bad release state, {0}, {1}".format(
                self.deployment_context.release_state,
                str(self.configuration.release_states),
            )
            raise PipelineViewError(message)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import pytest

from foodx_devops_tools.azure.cloud import (
    AzureCredentials,
    AzureSubscriptionConfiguration,
)


def pytest_addoption(parser):
    parser.addoption(
        "--subscription",
        action="store",
        default=None,
        help="Azure Cloud subscription id or name",
    )
    parser.addoption(
        "--tenant",
        action="store",
        default=None,
        help="Azure Cloud tenant id or name",
    )
    parser.addoption(
        "--userid",
        action="store",
        default=None,
        help="Azure Cloud service principal id",
    )
    parser.addoption(
        "--secret",
        action="store",
        default=None,
        help="Azure Cloud service principal secret",
    )
    parser.addoption(
        "--parameters",
        action="store",
        default=None,
        help="Path to deployment ARM template parameters file",
    )
    parser.addoption(
        "--template",
        action="store",
        default=None,
        help="Path to deployment ARM template file",
    )


@pytest.fixture
def user_subscription(request) -> AzureSubscriptionConfiguration:
    this_id = request.config.getoption("--subscription")
    this_tenant = request.config.getoption("--tenant")
    this_sub = AzureSubscriptionConfiguration(
        subscription_id=this_id, tenant_id=this_tenant
    )
    return this_sub


@pytest.fixture
def service_principal_credentials(request) -> AzureCredentials:
    this_id = request.config.getoption("--userid")
    this_password = request.config.getoption("--secret")
    this_sub = request.config.getoption("--subscription")
    this_tenant = request.config.getoption("--tenant")
    this_creds = AzureCredentials(
        userid=this_id,
        secret=this_password,
        subscription=this_sub,
        tenant=this_tenant,
        name="mock-name",
    )
    return this_creds


@pytest.fixture
def template_file(request) -> pathlib.Path:
    return pathlib.Path(request.config.getoption("--template"))


@pytest.fixture
def parameters_file(request) -> pathlib.Path:
    return pathlib.Path(request.config.getoption("--parameters"))



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import click

from foodx_devops_tools._declarations import (
    DEFAULT_CONSOLE_LOGGING_ENABLED,
    DEFAULT_FILE_LOGGING_DISABLED,
    DEFAULT_LOG_LEVEL,
    VALID_LOG_LEVELS,
)
from foodx_devops_tools._logging import LoggingState
from foodx_devops_tools._version import acquire_version

from .azure_cd import azure_subcommand
from .npm_ci import npm_subcommand

DEFAULT_LOG_FILE = pathlib.Path("puff_run.log")


@click.group()
@click.version_option(version=acquire_version())
@click.option(
    "--log-enable-console",
    "enable_console_log",
    default=DEFAULT_CONSOLE_LOGGING_ENABLED,
    help="Log to console.",
    is_flag=True,
)
@click.option(
    "--log-disable-file",
    "disable_file_log",
    default=DEFAULT_FILE_LOGGING_DISABLED,
    help="Disable file logging.",
    is_flag=True,
)
@click.option(
    "--log-level",
    "log_level",
    default=DEFAULT_LOG_LEVEL,
    help="Select logging level to apply to all enabled log sinks.",
    type=click.Choice(VALID_LOG_LEVELS, case_sensitive=False),
)
def release_flow(
    disable_file_log: bool, enable_console_log: bool, log_level: str
) -> None:
    """Release flow command group."""
    # currently no need to change logging configuration at run time,
    # so no need to preserve the object.
    LoggingState(
        disable_file_logging=disable_file_log,
        enable_console_logging=enable_console_log,
        log_level_text=log_level,
        default_log_file=DEFAULT_LOG_FILE,
    )


release_flow.add_command(azure_subcommand, name="azure")
release_flow.add_command(npm_subcommand, name="npm")



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy

import pytest

from foodx_devops_tools.pipeline_config import PipelineConfiguration
from tests.ci.support.pipeline_config import MOCK_RESULTS


@pytest.fixture()
def prep_frame_data(mock_flattened_deployment):
    pipeline_config = PipelineConfiguration.parse_obj(MOCK_RESULTS)
    frame_data = copy.deepcopy(pipeline_config.frames.frames["f1"])

    deployment_data = mock_flattened_deployment[0]
    deployment_data.data.iteration_context.append("f1")
    deployment_data.context.frame_name = "f1"
    deployment_data.context.release_state = "r1"
    deployment_data.context.azure_subscription_name = "sys1_c1_r1a"
    deployment_data.data.puff_map = pipeline_config.puff_map

    return deployment_data, frame_data



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.puff.arm import _linearize_name


class TestLinearizeName:
    MOCK_BASE = {
        "environments": dict(),
        "services": dict(),
    }

    def test_level0_name(self):
        expected_result = {
            "another-name": {
                "environments": dict(),
                "services": dict(),
            },
        }
        mock_file = "some-file"
        mock_input = {
            **self.MOCK_BASE,
            "name": "another-name",
        }

        result = _linearize_name(mock_input, mock_file)

        assert result == expected_result
        assert "name" not in result["another-name"]

    def test_level0_noname(self):
        expected_result = {
            "some-file": {
                "environments": dict(),
                "services": dict(),
            },
        }
        mock_file = "some-file"
        mock_input = self.MOCK_BASE.copy()

        result = _linearize_name(mock_input, mock_file)

        assert result == expected_result
        assert "name" not in result["some-file"]



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy
import enum
import logging

import pytest

from foodx_devops_tools.deploy_me._deployment import DeploymentState
from foodx_devops_tools.deploy_me._main import (
    ConfigurationPathsError,
    PipelineCliOptions,
    _report_results,
)
from foodx_devops_tools.deploy_me_entry import deploy_me
from tests.ci.support.click_runner import click_runner  # noqa: F401
from tests.ci.support.pipeline_config import (
    CLEAN_SPLIT,
    MOCK_SECRET,
    split_directories,
)


class TestReportResults:
    def test_success(self, capsys):
        mock_result = DeploymentState.ResultType.success
        number_iterations = 1

        _report_results(mock_result, number_iterations)

        captured = capsys.readouterr()
        assert "success: Deployment succeeded" in captured.out

    def test_skipped(self, capsys):
        mock_result = DeploymentState.ResultType.success
        number_iterations = 0

        _report_results(mock_result, number_iterations)

        captured = capsys.readouterr()
        assert "skipped: Deployment skipped" in captured.out

    def test_failed(self, capsys):
        mock_result = DeploymentState.ResultType.failed
        number_iterations = 1

        with pytest.raises(SystemExit):
            _report_results(mock_result, number_iterations)

        captured = capsys.readouterr()
        assert (
            "FAILED: Deployment failed. Check log for details" in captured.out
        )


@pytest.fixture()
def mock_getsha(mocker):
    def _apply(output: str = ""):
        mocker.patch(
            "foodx_devops_tools.deploy_me._main.get_sha",
            return_value=output,
        )

    return _apply


@pytest.fixture()
def mock_leakage_check(mocker):
    return mocker.patch(
        "foodx_devops_tools.deploy_me._main.check_credential_leakage"
    )


class TestDeployMe:
    EXPECTED_DEFAULT_OPTIONS = PipelineCliOptions(
        enable_validation=False,
        monitor_sleep_seconds=30,
        wait_timeout_seconds=(15 * 60),
    )

    class MockReleaseState(enum.Enum):
        r1 = enum.auto()

    def test_help(
        self,
        caplog,
        click_runner,
        mocker,
        mock_async_method,
        mock_getsha,
        mock_leakage_check,
    ):
        mock_input = [
            "--help",
        ]
        mock_getsha()
        mocker.patch(
            "foodx_devops_tools.deploy_me._main.identify_release_state",
            return_value=TestDeployMe.MockReleaseState.r1,
        )
        mock_deploy = mock_async_method(
            "foodx_devops_tools.deploy_me._main.do_deploy",
            return_value=DeploymentState(
                code=DeploymentState.ResultType.success
            ),
        )
        with split_directories(CLEAN_SPLIT.copy()) as (
            client_path,
            system_path,
        ):
            with caplog.at_level(logging.DEBUG):
                result = click_runner.invoke(
                    deploy_me, mock_input, input=MOCK_SECRET
                )

            assert result.exit_code == 0
            assert "Deploy system resources." in result.output

    def test_default(
        self,
        click_runner,
        caplog,
        mock_async_method,
        mock_getsha,
        mock_leakage_check,
        mocker,
    ):
        mock_input = list()

        result, mock_deploy = self._run_test(
            mock_input,
            caplog,
            click_runner,
            mock_async_method,
            mock_getsha,
            mocker,
        )

        assert result.exit_code == 0
        mock_deploy.assert_has_calls(
            [
                mocker.call(
                    mocker.ANY, mocker.ANY, self.EXPECTED_DEFAULT_OPTIONS
                ),
                mocker.call(
                    mocker.ANY, mocker.ANY, self.EXPECTED_DEFAULT_OPTIONS
                ),
            ]
        )
        mock_leakage_check.assert_called_once()

    def test_pipeline_id(
        self,
        click_runner,
        caplog,
        mock_async_method,
        mock_getsha,
        mock_leakage_check,
        mocker,
    ):
        mock_input = [
            "--pipeline-id",
            "refs/heads/main",
        ]

        result, mock_deploy = self._run_test(
            mock_input,
            caplog,
            click_runner,
            mock_async_method,
            mock_getsha,
            mocker,
        )

        assert result.exit_code == 0
        mock_deploy.assert_has_calls(
            [
                mocker.call(
                    mocker.ANY, mocker.ANY, self.EXPECTED_DEFAULT_OPTIONS
                ),
                mocker.call(
                    mocker.ANY, mocker.ANY, self.EXPECTED_DEFAULT_OPTIONS
                ),
            ]
        )

    def _run_test(
        self,
        mock_input,
        caplog,
        click_runner,
        mock_async_method,
        mock_getsha,
        mocker,
    ):
        mock_getsha()
        mocker.patch(
            "foodx_devops_tools.deploy_me._main.identify_release_state",
            return_value=TestDeployMe.MockReleaseState.r1,
        )
        mock_deploy = mock_async_method(
            "foodx_devops_tools.deploy_me._main.do_deploy",
            return_value=DeploymentState(
                code=DeploymentState.ResultType.success
            ),
        )
        with split_directories(CLEAN_SPLIT.copy()) as (
            client_path,
            system_path,
        ):
            mock_input += [
                str(client_path),
                str(system_path),
                "-",
                # --git-ref is mandatory, for now
                "--git-ref",
                "refs/heads/main",
            ]
            with caplog.at_level(logging.DEBUG):
                result = click_runner.invoke(
                    deploy_me, mock_input, input=MOCK_SECRET
                )

        return result, mock_deploy

    def test_validation(
        self,
        caplog,
        click_runner,
        mock_async_method,
        mock_getsha,
        mock_leakage_check,
        mocker,
    ):
        mock_input = [
            "--validation",
        ]

        result, mock_deploy = self._run_test(
            mock_input,
            caplog,
            click_runner,
            mock_async_method,
            mock_getsha,
            mocker,
        )

        assert result.exit_code == 0
        expected_options = copy.deepcopy(self.EXPECTED_DEFAULT_OPTIONS)
        expected_options.enable_validation = True
        mock_deploy.assert_has_calls(
            [
                mocker.call(mocker.ANY, mocker.ANY, expected_options),
                mocker.call(mocker.ANY, mocker.ANY, expected_options),
            ]
        )

    def test_monitor_sleep(
        self,
        click_runner,
        caplog,
        mock_async_method,
        mock_getsha,
        mock_leakage_check,
        mocker,
    ):
        mock_input = [
            "--monitor-sleep",
            "120",
        ]

        result, mock_deploy = self._run_test(
            mock_input,
            caplog,
            click_runner,
            mock_async_method,
            mock_getsha,
            mocker,
        )

        assert result.exit_code == 0
        expected_options = PipelineCliOptions(
            enable_validation=False,
            monitor_sleep_seconds=120,
            wait_timeout_seconds=(15 * 60),
        )
        mock_deploy.assert_has_calls(
            [
                mocker.call(mocker.ANY, mocker.ANY, expected_options),
                mocker.call(mocker.ANY, mocker.ANY, expected_options),
            ]
        )

    def test_wait_timeout(
        self,
        click_runner,
        caplog,
        mock_async_method,
        mock_getsha,
        mock_leakage_check,
        mocker,
    ):
        mock_input = [
            "--wait-timeout",
            "30",
        ]

        result, mock_deploy = self._run_test(
            mock_input,
            caplog,
            click_runner,
            mock_async_method,
            mock_getsha,
            mocker,
        )

        assert result.exit_code == 0
        expected_options = PipelineCliOptions(
            enable_validation=False,
            monitor_sleep_seconds=30,
            wait_timeout_seconds=(30 * 60),
        )
        mock_deploy.assert_has_calls(
            [
                mocker.call(mocker.ANY, mocker.ANY, expected_options),
                mocker.call(mocker.ANY, mocker.ANY, expected_options),
            ]
        )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.pipeline_config import PipelineConfigurationPaths
from foodx_devops_tools.pipeline_config.exceptions import (
    ConfigurationPathsError,
)
from tests.ci.support.pipeline_config import CLEAN_SPLIT, split_directories


class TestFromPaths:
    def test_clean(self):
        with split_directories(CLEAN_SPLIT.copy()) as (
            client_path,
            system_path,
        ):
            client_config = client_path / "configuration"
            system_config = system_path / "configuration"
            under_test = PipelineConfigurationPaths.from_paths(
                client_config, system_config
            )

            all_paths = set(client_config.iterdir()).union(
                set(system_config.iterdir())
            )
            all_names = CLEAN_SPLIT["client"].union(CLEAN_SPLIT["system"])
            for this_name in all_names:
                if this_name not in ["static_secrets", "context"]:
                    assert getattr(under_test, this_name) in all_paths

            assert {x.stem for x in under_test.static_secrets} == {
                "sys1_c1_r1a"
            }
            assert {x.stem for x in under_test.context} == {"my_key"}

    def test_no_secrets(self):
        split = {
            # duplicates between client and system dirs
            "client": {
                "clients",
                "deployments",
                "frames",
                "puff_map",
                "service_principals",
            },
            "system": {
                "release_states",
                "subscriptions",
                "systems",
                "tenants",
            },
        }
        with split_directories(split) as (
            client_path,
            system_path,
        ):
            client_config = client_path / "configuration"
            system_config = system_path / "configuration"
            under_test = PipelineConfigurationPaths.from_paths(
                client_config, system_config
            )

            all_paths = set(client_config.iterdir()).union(
                set(system_config.iterdir())
            )
            all_names = split["client"].union(split["system"])
            for this_name in all_names:
                assert getattr(under_test, this_name) in all_paths

            assert not under_test.static_secrets

    def test_duplicates_raises(self):
        split = {
            # duplicates between client and system dirs
            "client": {
                "clients",
                "deployments",
                "frames",
                "puff_map",
                "release_states",
                "service_principals",
            },
            "system": {
                "release_states",
                "subscriptions",
                "systems",
                "tenants",
            },
        }
        with split_directories(split) as (
            client_path,
            system_path,
        ), pytest.raises(
            ConfigurationPathsError,
            match=r"^Duplicate files between directories",
        ):
            client_config = client_path / "configuration"
            system_config = system_path / "configuration"
            PipelineConfigurationPaths.from_paths(client_config, system_config)



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import io
import sys


@contextlib.contextmanager
def capture_stdout_stderr(reset_seek: bool = True) -> None:
    new_stderr = io.StringIO()
    new_stdout = io.StringIO()
    with contextlib.redirect_stderr(
        new_stderr
    ) as err, contextlib.redirect_stdout(new_stdout) as out:
        yield out, err

    if reset_seek:
        # Reset the file pointer to the beginning of the stream.
        out.seek(0)
        err.seek(0)



#  Copyright (c) 2022-2022 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Application step action implementations."""

from ._delay import delay_step  # noqa: F401
from ._deploy import deploy_step  # noqa: F401
from ._script import script_step  # noqa: F401



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

from foodx_devops_tools.pipeline_config import PipelineConfigurationPaths


class TestPipelineConfigurationPaths:
    def test_template_context(self, mocker):
        mock_config = pathlib.Path("mock_config")
        mock_system = pathlib.Path("mock_system")
        mocker.patch(
            "foodx_devops_tools.pipeline_config"
            "._paths.PipelineConfigurationPaths"
            "._PipelineConfigurationPaths__acquire_static_secrets"
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config"
            "._paths.PipelineConfigurationPaths"
            "._PipelineConfigurationPaths__acquire_client_files"
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config"
            "._paths.PipelineConfigurationPaths"
            "._PipelineConfigurationPaths__acquire_system_files"
        )
        mock_paths = [
            {pathlib.Path("one")},
            {pathlib.Path("two")},
        ]
        mocker.patch(
            "foodx_devops_tools.pipeline_config"
            "._paths.PipelineConfigurationPaths"
            "._PipelineConfigurationPaths__acquire_subdir_files",
            side_effect=mock_paths,
        )

        under_test = PipelineConfigurationPaths.from_paths(
            mock_config, mock_system
        )

        expected_context = mock_paths[0].union(mock_paths[1])
        assert under_test.context == expected_context



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.puff.arm import _linearize_regions


class TestLinearizedRegions:
    def test_regions(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
        }
        expected_result = {
            "this.stub.r1": {
                "region": "r1",
                "k1": "e1r1k1",
            },
            "this.stub.r2": {
                "region": "r2",
                "k1": "bk1",
                "k2": "e1r2k2",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_multiple_stubs_mixed_regions(self):
        mock_base = {
            "stub1": {
                "k1": "stub1.k1",
                # "full" region in a stub
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
            "stub2": {
                "k1": "stub2.k1",
                # empty regions in a stub
                "regions": list(),
            },
        }
        expected_result = {
            "stub1.r1": {
                "region": "r1",
                "k1": "e1r1k1",
            },
            "stub1.r2": {
                "region": "r2",
                "k1": "stub1.k1",
                "k2": "e1r2k2",
            },
            "stub2": {
                "k1": "stub2.k1",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_empty_regions(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
                "regions": list(),
            },
        }
        expected_result = {
            "this.stub": {
                "k1": "bk1",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_empty_region(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
                "regions": [
                    {"r1": dict()},
                ],
            },
        }
        expected_result = {
            "this.stub.r1": {
                "region": "r1",
                "k1": "bk1",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_none_region(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
                "regions": [
                    {"r1": None},
                ],
            },
        }
        expected_result = {
            "this.stub.r1": {
                "region": "r1",
                "k1": "bk1",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_no_regions(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
            },
        }
        expected_result = {
            "this.stub": {
                "k1": "bk1",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_regions_empty_base(self):
        mock_base = {
            "this.stub": {
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
        }
        expected_result = {
            "this.stub.r1": {
                "region": "r1",
                "k1": "e1r1k1",
            },
            "this.stub.r2": {
                "region": "r2",
                "k2": "e1r2k2",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_no_regions_empty_base(self):
        mock_base = {
            "this.stub": dict(),
        }
        expected_result = {
            "this.stub": dict(),
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_merged_regions(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
                "regions": [
                    # regions merged from other iterables (services or
                    # environments).
                    {"r1": {"k1": "e1r1k1"}},
                    {"r1": {"k2": "e2r1k2"}},
                    {"r1": {"k3": "e3r1k3"}},
                    {"r2": {"k4": "e1r2k2"}},
                    {"r2": {"k5": "e2r2k3"}},
                ],
            },
        }
        expected_result = {
            "this.stub.r1": {
                "region": "r1",
                "k1": "e1r1k1",
                "k2": "e2r1k2",
                "k3": "e3r1k3",
            },
            "this.stub.r2": {
                "region": "r2",
                "k1": "bk1",
                "k4": "e1r2k2",
                "k5": "e2r2k3",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result

    def test_none_regions(self):
        mock_base = {
            "this.stub": {
                "k1": "bk1",
                "regions": [
                    None,
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
        }
        expected_result = {
            "this.stub.r2": {
                "region": "r2",
                "k1": "bk1",
                "k2": "e1r2k2",
            },
        }
        result = _linearize_regions(mock_base)

        assert result == expected_result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib
import typing
import uuid

import pytest
from ruamel.yaml import YAML

from foodx_devops_tools.puff.arm import (
    ArmTemplateError,
    do_arm_template_parameter_action,
)

from .support import initialize_filesystem


class TestGenerateArmTemplateParameters:
    @pytest.mark.asyncio
    async def test_file_logging(self, capsys, mocker):
        puff_content = """---
environments:
  e1:
    k1: e1k1
    k2: e1k2
"""
        this_yaml = YAML(typ="safe")
        data = this_yaml.load(puff_content)
        mocker.patch("foodx_devops_tools.puff.arm.load_yaml", return_value=data)
        mocker.patch("foodx_devops_tools.puff.arm._save_parameter_file")
        this_directory = pathlib.Path("some/path")
        puff_path = this_directory / "this_file.yml"

        await do_arm_template_parameter_action(puff_path, None, False, False)

        out = capsys.readouterr().out

        assert "loading, {0}".format(str(puff_path)) in out

    @pytest.mark.asyncio
    async def test_name_optional(self, mocker):
        puff_content = """---
environments:
  e1:
    k1: e1k1
    k2: e1k2
  e2:
    k1: e2k1
    k2: e2k2
"""
        expected_files = [
            "this_file.e1.json",
            "this_file.e2.json",
        ]
        expected_content = [
            {
                "k1": "e1k1",
                "k2": "e1k2",
            },
            {
                "k1": "e2k1",
                "k2": "e2k2",
            },
        ]

        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    async def _do_file_test(
        self,
        expected_files: typing.List[str],
        expected_values: typing.List[dict],
        puff_content: str,
        this_mocker,
        is_delete_action: bool = False,
    ) -> None:
        this_yaml = YAML(typ="safe")
        data = this_yaml.load(puff_content)
        this_mocker.patch(
            "foodx_devops_tools.puff.arm.load_yaml", return_value=data
        )
        mock_save = this_mocker.patch(
            "foodx_devops_tools.puff.arm._save_parameter_file"
        )
        this_directory = pathlib.Path("some/path")
        puff_path = this_directory / "this_file.yml"

        await do_arm_template_parameter_action(
            puff_path, None, is_delete_action, False
        )

        expected_paths = [this_directory / x for x in expected_files]

        mock_save.has_calls(
            [
                this_mocker.call(expected_paths[x], expected_values[x])
                for x in range(0, len(expected_paths))
            ]
        )

    @pytest.mark.asyncio
    async def test_environments_regions_clean(self, mocker):
        puff_content = """---
name: some-name
environments:
  e1:
    k1: e1k1
    k2: e1k2
  e2:
    k1: e2k1
    k2: e2k2
    regions:
      - r1:
          k3: e2r1k3
"""
        expected_files = [
            "some-name.e1.json",
            "some-name.e2.r1.json",
        ]
        expected_content = [
            {
                "k1": "e1k1",
                "k2": "e1k2",
            },
            {
                "k1": "e2k1",
                "k2": "e2k2",
                "k3": "e2r1k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_services_clean(self, mocker):
        puff_content = """---
name: some-name
services:
  s1:
    k1: s1k1
    k2: s1k2
  s2:
    k1: s2k1
    k2: s2k2
    k3: s2k3
"""
        expected_files = [
            (".".join(["some-name", x, "json"])) for x in ["s1", "s2"]
        ]
        expected_content = [
            {
                "k1": "s1k1",
                "k2": "s1k2",
            },
            {
                "k1": "s2k1",
                "k2": "s2k2",
                "k3": "s2k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_environments_clean(self, mocker):
        puff_content = """---
name: some-name
environments:
  e1:
    k1: e1k1
    k2: e1k2
  e2:
    k1: e2k1
    k2: e2k2
    k3: e2k3
"""
        expected_files = [
            (".".join(["some-name", x, "json"])) for x in ["e1", "e2"]
        ]
        expected_content = [
            {
                "k1": "e1k1",
                "k2": "e1k2",
            },
            {
                "k1": "e2k1",
                "k2": "e2k2",
                "k3": "e2k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_regions_clean(self, mocker):
        puff_content = """---
name: some-name
environments:
  e1:
    regions:
      - r1:
          k1: e1r1k1
          k2: e1r1k2
      - r2:
          k1: e1r2k1
          k2: e1r2k2
          k3: e1r2k3
"""
        expected_files = [
            (".".join(["some-name", x, "json"])) for x in ["e1.r1", "e1.r2"]
        ]
        expected_content = [
            {
                "k1": "e1r1k1",
                "k2": "e1r1k2",
            },
            {
                "k1": "e1r2k1",
                "k2": "e1r2k2",
                "k3": "e1r2k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_all_in_clean(self, mocker):
        puff_content = """---
name: some-name
k1: bk1
k2: bk2
environments:
  e1:
    k3: e1k3
  e2:
    k3: e2k3
    regions:
      - r2:
          k1: e2r2k1
services:
  s1:
    k1: s1k1
    k2: s1k2
    environments:
      e1:
        k5: s1e1k5
        regions:
          - r1:
              k2: s1e1r1k2
              k6: s1e1r1k6
  s2:
    k1: s2k1
    k2: s2k2
    k3: s2k3
"""
        expected_files = [
            (".".join(["some-name", x, "json"]))
            for x in ["s1.e1.r1", "s1.e2.r2", "s2.e1", "s2.e2.r2"]
        ]
        expected_content = [
            {
                "k1": "value1",
                "k2": "s1e1r1k2",
                "k3": "e1k3",
                "k5": "s1e1k5",
                "k6": "s1e1r1k6",
            },
            {
                "k1": "e2r2k1",
                "k2": "bk2",
                "k3": "e2k3",
            },
            {
                "k1": "s2k1",
                "k2": "s2k2",
                "k3": "e1k3",
            },
            {
                "k1": "e2r2k1",
                "k2": "s2k2",
                "k3": "e2k3",
            },
        ]
        await self._do_file_test(
            expected_files,
            expected_content,
            puff_content,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_empty_raises(self, tmp_path_factory):
        puff_content = ""
        this_directory = tmp_path_factory.mktemp(str(uuid.uuid4())).absolute()
        puff_path = this_directory / "this_file.yml"
        initialize_filesystem(puff_path, puff_content)

        with pytest.raises(ArmTemplateError):
            await do_arm_template_parameter_action(
                puff_path, None, False, False
            )

    @pytest.mark.asyncio
    async def test_empty_services_raises(self, tmp_path_factory):
        puff_content = """---
environments:
  e1:
    v: v
services:
"""
        this_directory = tmp_path_factory.mktemp(str(uuid.uuid4())).absolute()
        puff_path = this_directory / "this_file.yml"
        initialize_filesystem(puff_path, puff_content)

        with pytest.raises(ArmTemplateError):
            await do_arm_template_parameter_action(
                puff_path, None, False, False
            )
